# coding: utf-8

"""
    FlexPrice API

    FlexPrice API Service

    The version of the OpenAPI document: 1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, StrictBool, StrictInt, StrictStr
from typing import Any, ClassVar, Dict, List, Optional
from flexprice.models.types_billing_period import TypesBillingPeriod
from flexprice.models.types_feature_type import TypesFeatureType
from typing import Optional, Set
from typing_extensions import Self

class DtoCreateEntitlementRequest(BaseModel):
    """
    DtoCreateEntitlementRequest
    """ # noqa: E501
    feature_id: StrictStr
    feature_type: TypesFeatureType
    is_enabled: Optional[StrictBool] = None
    is_soft_limit: Optional[StrictBool] = None
    plan_id: Optional[StrictStr] = None
    static_value: Optional[StrictStr] = None
    usage_limit: Optional[StrictInt] = None
    usage_reset_period: Optional[TypesBillingPeriod] = None
    __properties: ClassVar[List[str]] = ["feature_id", "feature_type", "is_enabled", "is_soft_limit", "plan_id", "static_value", "usage_limit", "usage_reset_period"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of DtoCreateEntitlementRequest from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of DtoCreateEntitlementRequest from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "feature_id": obj.get("feature_id"),
            "feature_type": obj.get("feature_type"),
            "is_enabled": obj.get("is_enabled"),
            "is_soft_limit": obj.get("is_soft_limit"),
            "plan_id": obj.get("plan_id"),
            "static_value": obj.get("static_value"),
            "usage_limit": obj.get("usage_limit"),
            "usage_reset_period": obj.get("usage_reset_period")
        })
        return _obj





# coding: utf-8

# flake8: noqa
"""
    FlexPrice API

    FlexPrice API Service

    The version of the OpenAPI document: 1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


# import models into model package
from flexprice.models.dto_address import DtoAddress
from flexprice.models.dto_aggregated_entitlement import DtoAggregatedEntitlement
from flexprice.models.dto_aggregated_feature import DtoAggregatedFeature
from flexprice.models.dto_auth_response import DtoAuthResponse
from flexprice.models.dto_billing_period_info import DtoBillingPeriodInfo
from flexprice.models.dto_bulk_ingest_event_request import DtoBulkIngestEventRequest
from flexprice.models.dto_create_api_key_request import DtoCreateAPIKeyRequest
from flexprice.models.dto_create_api_key_response import DtoCreateAPIKeyResponse
from flexprice.models.dto_create_customer_request import DtoCreateCustomerRequest
from flexprice.models.dto_create_entitlement_request import DtoCreateEntitlementRequest
from flexprice.models.dto_create_environment_request import DtoCreateEnvironmentRequest
from flexprice.models.dto_create_feature_request import DtoCreateFeatureRequest
from flexprice.models.dto_create_integration_request import DtoCreateIntegrationRequest
from flexprice.models.dto_create_invoice_line_item_request import DtoCreateInvoiceLineItemRequest
from flexprice.models.dto_create_invoice_request import DtoCreateInvoiceRequest
from flexprice.models.dto_create_meter_request import DtoCreateMeterRequest
from flexprice.models.dto_create_payment_request import DtoCreatePaymentRequest
from flexprice.models.dto_create_plan_entitlement_request import DtoCreatePlanEntitlementRequest
from flexprice.models.dto_create_plan_price_request import DtoCreatePlanPriceRequest
from flexprice.models.dto_create_plan_request import DtoCreatePlanRequest
from flexprice.models.dto_create_price_request import DtoCreatePriceRequest
from flexprice.models.dto_create_price_tier import DtoCreatePriceTier
from flexprice.models.dto_create_subscription_request import DtoCreateSubscriptionRequest
from flexprice.models.dto_create_task_request import DtoCreateTaskRequest
from flexprice.models.dto_create_tenant_request import DtoCreateTenantRequest
from flexprice.models.dto_create_wallet_request import DtoCreateWalletRequest
from flexprice.models.dto_customer_entitlements_response import DtoCustomerEntitlementsResponse
from flexprice.models.dto_customer_invoice_summary import DtoCustomerInvoiceSummary
from flexprice.models.dto_customer_multi_currency_invoice_summary import DtoCustomerMultiCurrencyInvoiceSummary
from flexprice.models.dto_customer_response import DtoCustomerResponse
from flexprice.models.dto_customer_usage_summary_response import DtoCustomerUsageSummaryResponse
from flexprice.models.dto_entitlement_response import DtoEntitlementResponse
from flexprice.models.dto_entitlement_source import DtoEntitlementSource
from flexprice.models.dto_environment_response import DtoEnvironmentResponse
from flexprice.models.dto_event import DtoEvent
from flexprice.models.dto_feature_response import DtoFeatureResponse
from flexprice.models.dto_feature_usage_summary import DtoFeatureUsageSummary
from flexprice.models.dto_get_events_response import DtoGetEventsResponse
from flexprice.models.dto_get_preview_invoice_request import DtoGetPreviewInvoiceRequest
from flexprice.models.dto_get_usage_by_meter_request import DtoGetUsageByMeterRequest
from flexprice.models.dto_get_usage_by_subscription_request import DtoGetUsageBySubscriptionRequest
from flexprice.models.dto_get_usage_by_subscription_response import DtoGetUsageBySubscriptionResponse
from flexprice.models.dto_get_usage_request import DtoGetUsageRequest
from flexprice.models.dto_get_usage_response import DtoGetUsageResponse
from flexprice.models.dto_ingest_event_request import DtoIngestEventRequest
from flexprice.models.dto_invoice_line_item_response import DtoInvoiceLineItemResponse
from flexprice.models.dto_invoice_response import DtoInvoiceResponse
from flexprice.models.dto_linked_integrations_response import DtoLinkedIntegrationsResponse
from flexprice.models.dto_list_customers_response import DtoListCustomersResponse
from flexprice.models.dto_list_entitlements_response import DtoListEntitlementsResponse
from flexprice.models.dto_list_environments_response import DtoListEnvironmentsResponse
from flexprice.models.dto_list_features_response import DtoListFeaturesResponse
from flexprice.models.dto_list_invoices_response import DtoListInvoicesResponse
from flexprice.models.dto_list_payments_response import DtoListPaymentsResponse
from flexprice.models.dto_list_plans_response import DtoListPlansResponse
from flexprice.models.dto_list_prices_response import DtoListPricesResponse
from flexprice.models.dto_list_secrets_response import DtoListSecretsResponse
from flexprice.models.dto_list_subscription_pauses_response import DtoListSubscriptionPausesResponse
from flexprice.models.dto_list_subscriptions_response import DtoListSubscriptionsResponse
from flexprice.models.dto_list_tasks_response import DtoListTasksResponse
from flexprice.models.dto_list_wallet_transactions_response import DtoListWalletTransactionsResponse
from flexprice.models.dto_login_request import DtoLoginRequest
from flexprice.models.dto_meter_response import DtoMeterResponse
from flexprice.models.dto_pause_subscription_request import DtoPauseSubscriptionRequest
from flexprice.models.dto_payment_attempt_response import DtoPaymentAttemptResponse
from flexprice.models.dto_payment_response import DtoPaymentResponse
from flexprice.models.dto_plan_response import DtoPlanResponse
from flexprice.models.dto_price_response import DtoPriceResponse
from flexprice.models.dto_resume_subscription_request import DtoResumeSubscriptionRequest
from flexprice.models.dto_secret_response import DtoSecretResponse
from flexprice.models.dto_sign_up_request import DtoSignUpRequest
from flexprice.models.dto_subscription_pause_response import DtoSubscriptionPauseResponse
from flexprice.models.dto_subscription_response import DtoSubscriptionResponse
from flexprice.models.dto_subscription_usage_by_meters_response import DtoSubscriptionUsageByMetersResponse
from flexprice.models.dto_task_response import DtoTaskResponse
from flexprice.models.dto_tenant_billing_details import DtoTenantBillingDetails
from flexprice.models.dto_tenant_billing_usage import DtoTenantBillingUsage
from flexprice.models.dto_tenant_response import DtoTenantResponse
from flexprice.models.dto_top_up_wallet_request import DtoTopUpWalletRequest
from flexprice.models.dto_update_customer_request import DtoUpdateCustomerRequest
from flexprice.models.dto_update_entitlement_request import DtoUpdateEntitlementRequest
from flexprice.models.dto_update_environment_request import DtoUpdateEnvironmentRequest
from flexprice.models.dto_update_feature_request import DtoUpdateFeatureRequest
from flexprice.models.dto_update_meter_request import DtoUpdateMeterRequest
from flexprice.models.dto_update_payment_request import DtoUpdatePaymentRequest
from flexprice.models.dto_update_payment_status_request import DtoUpdatePaymentStatusRequest
from flexprice.models.dto_update_plan_entitlement_request import DtoUpdatePlanEntitlementRequest
from flexprice.models.dto_update_plan_price_request import DtoUpdatePlanPriceRequest
from flexprice.models.dto_update_plan_request import DtoUpdatePlanRequest
from flexprice.models.dto_update_price_request import DtoUpdatePriceRequest
from flexprice.models.dto_update_task_status_request import DtoUpdateTaskStatusRequest
from flexprice.models.dto_update_tenant_request import DtoUpdateTenantRequest
from flexprice.models.dto_update_wallet_request import DtoUpdateWalletRequest
from flexprice.models.dto_usage_result import DtoUsageResult
from flexprice.models.dto_user_response import DtoUserResponse
from flexprice.models.dto_wallet_balance_response import DtoWalletBalanceResponse
from flexprice.models.dto_wallet_response import DtoWalletResponse
from flexprice.models.dto_wallet_transaction_response import DtoWalletTransactionResponse
from flexprice.models.errors_error_detail import ErrorsErrorDetail
from flexprice.models.errors_error_response import ErrorsErrorResponse
from flexprice.models.meter_aggregation import MeterAggregation
from flexprice.models.meter_filter import MeterFilter
from flexprice.models.price_jsonb_transform_quantity import PriceJSONBTransformQuantity
from flexprice.models.price_price import PricePrice
from flexprice.models.price_price_tier import PricePriceTier
from flexprice.models.price_transform_quantity import PriceTransformQuantity
from flexprice.models.subscription_subscription_line_item import SubscriptionSubscriptionLineItem
from flexprice.models.subscription_subscription_pause import SubscriptionSubscriptionPause
from flexprice.models.types_aggregation_type import TypesAggregationType
from flexprice.models.types_auto_topup_trigger import TypesAutoTopupTrigger
from flexprice.models.types_billing_cadence import TypesBillingCadence
from flexprice.models.types_billing_model import TypesBillingModel
from flexprice.models.types_billing_period import TypesBillingPeriod
from flexprice.models.types_billing_tier import TypesBillingTier
from flexprice.models.types_entity_type import TypesEntityType
from flexprice.models.types_feature_type import TypesFeatureType
from flexprice.models.types_file_type import TypesFileType
from flexprice.models.types_invoice_billing_reason import TypesInvoiceBillingReason
from flexprice.models.types_invoice_cadence import TypesInvoiceCadence
from flexprice.models.types_invoice_status import TypesInvoiceStatus
from flexprice.models.types_invoice_type import TypesInvoiceType
from flexprice.models.types_pagination_response import TypesPaginationResponse
from flexprice.models.types_pause_mode import TypesPauseMode
from flexprice.models.types_pause_status import TypesPauseStatus
from flexprice.models.types_payment_destination_type import TypesPaymentDestinationType
from flexprice.models.types_payment_method_type import TypesPaymentMethodType
from flexprice.models.types_payment_status import TypesPaymentStatus
from flexprice.models.types_price_type import TypesPriceType
from flexprice.models.types_reset_usage import TypesResetUsage
from flexprice.models.types_resume_mode import TypesResumeMode
from flexprice.models.types_secret_provider import TypesSecretProvider
from flexprice.models.types_secret_type import TypesSecretType
from flexprice.models.types_status import TypesStatus
from flexprice.models.types_subscription_status import TypesSubscriptionStatus
from flexprice.models.types_task_status import TypesTaskStatus
from flexprice.models.types_task_type import TypesTaskType
from flexprice.models.types_transaction_reason import TypesTransactionReason
from flexprice.models.types_transaction_status import TypesTransactionStatus
from flexprice.models.types_transaction_type import TypesTransactionType
from flexprice.models.types_wallet_config import TypesWalletConfig
from flexprice.models.types_wallet_config_price_type import TypesWalletConfigPriceType
from flexprice.models.types_wallet_status import TypesWalletStatus
from flexprice.models.types_wallet_tx_reference_type import TypesWalletTxReferenceType
from flexprice.models.types_wallet_type import TypesWalletType
from flexprice.models.types_window_size import TypesWindowSize




import asyncio
import threading
import logging
from typing import Dict, Any, Optional, List, Union, Callable
from queue import Queue, Empty
import time
import random
import traceback

logger = logging.getLogger(__name__)

class AsyncEventProcessor:
    """Handler for asynchronous event processing with retries."""
    
    def __init__(self, 
                 max_retries: int = 3, 
                 retry_delay: float = 1.0,
                 max_queue_size: int = 1000,
                 workers: int = 2):
        """
        Initialize the async event processor.
        
        Args:
            max_retries: Maximum number of retry attempts for failed requests
            retry_delay: Base delay between retries (exponential backoff is applied)
            max_queue_size: Maximum number of events to queue
            workers: Number of worker threads to process events
        """
        self.max_retries = max_retries
        self.retry_delay = retry_delay
        self.event_queue = Queue(maxsize=max_queue_size)
        self.workers = workers
        self._worker_threads = []
        self._running = False
        
    def start(self):
        """Start the event processing workers."""
        if self._running:
            return
            
        self._running = True
        for i in range(self.workers):
            thread = threading.Thread(
                target=self._process_queue,
                name=f"event-worker-{i}",
                daemon=True
            )
            thread.start()
            self._worker_threads.append(thread)
        
        logger.info(f"Started {self.workers} async event processor workers")
    
    def stop(self):
        """Stop the event processing."""
        self._running = False
        for thread in self._worker_threads:
            thread.join(timeout=1.0)
        self._worker_threads = []
        
    def _process_queue(self):
        """Worker thread function to process events from the queue."""
        while self._running:
            try:
                event_data = self.event_queue.get(block=True, timeout=0.5)
                if event_data is None:
                    continue
                
                # Validate event_data structure
                if not isinstance(event_data, tuple) or len(event_data) != 3:
                    logger.error(f"Invalid event data format: {event_data}")
                    self.event_queue.task_done()
                    continue
                    
                event, api_instance, callback = event_data
                self._process_event(event, api_instance, callback)
                self.event_queue.task_done()
            except Empty:
                # Empty queue is normal, just continue polling
                pass
            except Exception as e:
                if not isinstance(e, asyncio.TimeoutError):
                    error_details = traceback.format_exc()
                    logger.error(f"Error in event processor: {str(e)}\n{error_details}")
    
    def _process_event(self, event, api_instance, callback):
        """Process a single event with retries."""
        attempt = 0
        last_error = None
        
        while attempt <= self.max_retries:
            try:
                if attempt > 0:
                    # Exponential backoff with jitter
                    delay = self.retry_delay * (2 ** (attempt - 1)) 
                    delay = delay * (0.5 + random.random())
                    time.sleep(delay)
                    
                logger.debug(f"Sending event (attempt {attempt+1}/{self.max_retries+1})")
                result = api_instance.events_post(event=event)
                
                if callback:
                    callback(result=result, error=None, success=True)
                
                logger.info(f"Successfully sent event to FlexPrice API")
                return
            except Exception as e:
                last_error = e
                error_details = traceback.format_exc()
                logger.warning(f"Event delivery failed (attempt {attempt+1}/{self.max_retries+1}): {str(e)}\n{error_details}")
                attempt += 1
                
        # All retries failed
        logger.error(f"Event delivery permanently failed after {self.max_retries+1} attempts: {str(last_error)}")
        if callback:
            callback(result=None, error=last_error, success=False)
    
    def submit_event(self, event, api_instance, callback=None):
        """
        Submit an event for asynchronous processing.
        
        Args:
            event: The event data to process
            api_instance: The API client instance to use for processing
            callback: Optional callback function that receives (result, error, success) params
            
        Returns:
            bool: True if the event was queued, False if the queue is full
        """
        try:
            # Validate inputs
            if event is None:
                logger.error("Cannot submit None event")
                return False
            if api_instance is None:
                logger.error("Cannot submit event with None api_instance")
                return False
                
            self.event_queue.put_nowait((event, api_instance, callback))
            logger.debug(f"Event queued for async processing")
            return True
        except Exception as e:
            logger.error(f"Failed to queue event: {str(e)}\n{traceback.format_exc()}")
            return False

# Global processor instance
_processor = AsyncEventProcessor()
_processor.start()

def submit_event_async(event, api_instance, callback=None):
    """
    Submit an event for asynchronous processing with automatic retries.
    
    Args:
        event: The event data to process
        api_instance: The API client instance to use
        callback: Optional callback function with signature (result, error, success)
        
    Returns:
        bool: True if the event was queued successfully, False otherwise
    """
    return _processor.submit_event(event, api_instance, callback)

# Ensure the processor is stopped when the program exits
import atexit
atexit.register(_processor.stop)
